// Copyright (c) 2012, Ladislav Thon. All rights reserved. Use of this source
// code is governed by a BSD-style license that can be found in the LICENSE file.

library range;

class _RangeIterator implements Iterator<int> {
  final int _start, _end, _step;
  int _current;
  bool _finished = false;

  _RangeIterator(this._start, this._end, this._step);

  bool moveNext() {
    if (_current == null && !_finished) {
      _current = _start;
      return true;
    }

    _current += _step;
    if ((_start <= _end && _current > _end) || (_start > _end && _current < _end)) {
      _finished = true;
      return false;
    }

    return true;
  }

  int get current => _finished ? null : _current;
}

/// A range of [int]s that iterates from specified number to another specified
/// number, both of them inclusive, with an optional step (defaults to 1).
/// It can either iterate upwards (with positive step) or downwards (with
/// negative step).
class Range extends Iterable<int> {
  final int start, end, step;

  Range._(this.start, this.end, this.step);

  /// If [start] is `<` than [end], then [step] must be `> 0`.
  /// If [start] is `>` than [end], then [step] must be `< 0`.
  /// If [start] is `==` to [end], then [step] is ignored.
  /// If [step] is omitted, it defaults to `1` when `start < end`
  /// and to `-1` when `start > end`.
  factory Range(int start, int end, {int step}) {
    if (start == end) {
      step = 1; // doesn't really matter (even step of 0 is acceptable in this case)
    }

    if (step == 0) {
      throw new ArgumentError("Step must not be 0");
    }

    bool stepMustBePositive = start <= end;

    if (step == null) {
      step = stepMustBePositive ? 1 : -1;
    }

    bool stepIsPositive = step > 0;
    if (stepIsPositive != stepMustBePositive) {
      throw new ArgumentError("Step ${stepMustBePositive ? 'must be positive' : 'must be negative'} but is $step");
    }

    return new Range._(start, end, step);
  }

  Iterator<int> get iterator => new _RangeIterator(start, end, step);

  // optimized implementations of some Iterable methods
  bool contains(int number) => start <= number && number <= end;
  int get length => (end - start) ~/ step + 1;
  bool get isEmpty => length == 0;
}

/// Ranges from `start` inclusive to `end` exclusive/inclusive (depends
/// on the value of `inclusive` parameter) with step of 1. I.e., it's the
/// same as `new Range(start, end - 1)` or `new Range(start, end)`.
/// It doesn't support ranging downwards (i.e., `start > end`).
Range range(int start, int end, {bool inclusive: false})
    => new Range(start, inclusive ? end : end - 1, step: 1);
